/*
 * samsung.c:
 *
 * Copyright (c) 2000 James McKenzie <james@fishsoup.dhs.org>.
 * All rights reserved.
 *
 *  By using this file, you agree to the terms and conditions set
 *  forth in the LICENCE file which can be found at the top level directory.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Log: samsung.c,v $
 * Revision 1.2  2000/02/17 15:35:18  root
 * #
 *
 * Revision 1.1  2000/02/17 15:28:55  root
 * #
 *
 *
 */

static char rcsid[] =
  "$Id: samsung.c,v 1.2 2000/02/17 15:35:18 root Exp root $";

#define error_dialog( A ) fprintf(stderr, A )
#define update_status( A )
#define update_progress( A )

char *serial_port = "/dev/ttyS0";

/*
   ** Copyright (C) 2000 James McKenzie
   **
   ** Filename:    samsung.c
   ** Author:      James McKenzie
   ** Date:        12 January 1999
   ** Version:     1.1
   ** Platform:    Linux
   ** WWW ref:     http://gimel.esc.cam.ac.uk/~james/digimax.html
   ** Licence:     GNU Public Licence
   **
   ** This program is free software; you can redistribute it and/or modify
   ** it under the terms of the GNU General Public License as published by
   ** the Free Software Foundation; version 2 dated June, 1991.
   **
   ** This program is distributed in the hope that it will be useful,
   ** but WITHOUT ANY WARRANTY; without even the implied warranty of
   ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   ** GNU General Public License for more details.
   **
   ** You should have received a copy of the GNU General Public License
   ** along with this program;  if not, write to the Free Software
   ** Foundation, Inc., 675 Mass Ave., Cambridge, MA 02139, USA.
   **
   ** For more infomation see http://www.gnu.org/copyleft/gpl.html
   **
   ** Description:
   **  This program downloads images from a Samsung Digimax 800k
   **
   ** Todo list:
   **   - fix crc code
   **
 */

#include<stdio.h>
#include<sys/types.h>
#include<sys/stat.h>
#include<fcntl.h>
#include<unistd.h>
#include<termios.h>
#include<sys/ioctl.h>
#include<setjmp.h>
#include<sys/time.h>
#include<signal.h>

/* print all data transmitted and received via the serial port */
/* #define SDSC_TRACE */


struct Image
{
  void *image;
  char image_type[80];
  int image_info_size;
  int image_size;
};

typedef unsigned char byte;

const unsigned char SDSC_EOF = 0x04; //Eof (received, new)
const unsigned char SDSC_BINARY = 0x06; //Set binary mode, next block
const unsigned char SDSC_RETRANSMIT = 0x15; //Request re-transmittion
const unsigned char SDSC_GIMME = 0x42; //Request image (new) 
const unsigned char SDSC_START = 0x43; //Request transmittion
const unsigned char SDSC_GOODBYE = 0x45; //Reset state and disconnect
const unsigned char SDSC_NEXT = 0x53; //Next image
const unsigned char SDSC_HELLO = 0x57; //New proto 2
const unsigned char SDSC_PROTO2OK = 0x49; //Proto 2 ok

/* number of image data bytes transfered in each packet */
const int SDSC_BLOCKSIZE = 0x405;
const int SDSC_INFOSIZE = 0x85;
const int SDSC_TIMEOUT = 500000 /* usec */ ;
const int SDSC_RETRIES = 3;

#define DEBUG


int fd;

#define PROTO_UNKN   0
#define PROTO_PROTO1 1
#define PROTO_PROTO2 2

int proto = PROTO_UNKN;

static int
sendcommand (byte cmd)
{
  int i;

#ifdef DEBUG
  fprintf (stderr, "Sending byte 0x%02x\n", cmd);
#endif
  return write (fd, &cmd, 1);
}

static int
waitforinput ()
{
  fd_set rfds;
  struct timeval tv;

  tv.tv_sec = 0;
  tv.tv_usec = SDSC_TIMEOUT;

  FD_ZERO (&rfds);
  FD_SET (fd, &rfds);

  if (!select (fd + 1, &rfds, NULL, NULL, &tv))
    return 0;

  if (!FD_ISSET (fd, &rfds))
    return 0;

  return 1;
}

static int
checkcrc (byte * buf, int len)
{
/* FIXME: */
  return 1;
}

static int
receive (byte * buf, int length, int expected, int retry)
{
  int tries = 0;
  int n, total, i;
  byte *ptr;


  while (1)
    {

      ptr = buf;
      i = length;
      total = 0;

      while (waitforinput (fd))
        {

          n = read (fd, ptr, i);

          if (n > 0)
            {
              ptr += n;
              i -= n;
              total += n;

              if (checkcrc (buf, total) && (total == expected))
                {
#ifdef DEBUG
                  fprintf (stderr, "Read %d bytes as expected\n", total);
#endif
                  return total;

                }

            }
        }

#ifdef DEBUG
      fprintf (stderr, "Read %d byes of %d bytes expected\n", total,
               expected);
#endif

      if (!retry)
        return total;

      /*Ok didn't match expected - two posibities */
      /*1st we missed packet/crcerror */
      /*2nd this is the last packet */

      tries++;


      if ((total == 1) && (tries == 2) && (expected == SDSC_BLOCKSIZE))
        {
#ifdef DEBUG
          fprintf (stderr, "Detected EOF condition\n");
#endif
          return total;
        }

      if (tries == SDSC_RETRIES)
        {
#ifdef DEBUG
          fprintf (stderr, "Giving up\n");
#endif
          return 0;
        }

#ifdef DEBUG
      fprintf (stderr, "Retrying\n");
#endif
      sendcommand (SDSC_RETRANSMIT);


    }

}


static int
init_comms ()
{
  struct termios tios;
  unsigned char buf[0x100];
  int length;

  tcgetattr (fd, &tios);
  tios.c_cflag &= ~(CSIZE | CSTOPB | PARENB | CLOCAL | CRTSCTS);
  tios.c_cflag |= CLOCAL;
  tios.c_cflag |= CS8;

  tios.c_cflag |= CREAD | HUPCL;
  tios.c_iflag = IGNBRK | IGNPAR;

  tios.c_oflag = 0;
  tios.c_lflag = 0;
  tios.c_cc[VMIN] = 1;
  tios.c_cc[VTIME] = 0;


  cfsetospeed (&tios, B115200);
  cfsetispeed (&tios, B115200);

  if (tcsetattr (fd, TCSAFLUSH, &tios) < 0)
    error_dialog ("Error setting communication parameters");

  return 0;
}


static int
isnullinfo (byte * buf)
{
  int n = 128;
  buf += 3;

  while (n--)
    {
      if (*(buf++))
        return 0;
    }

  return 1;
}



static int
reset_state (void)
{
  unsigned char buffer[2048];
  int len;

  switch (proto)
    {
    case PROTO_PROTO2:
    case PROTO_PROTO1:
      do
        {
          sendcommand (SDSC_NEXT);
          sendcommand (SDSC_START); /*Read out a header */
          len = receive (buffer, sizeof (buffer), SDSC_INFOSIZE, 1);

          if (len != SDSC_INFOSIZE)
            return 0;
        }
      while (!isnullinfo (buffer));
      sendcommand (SDSC_NEXT);  // One more to disconnect
      return 1;
    default:
      //sendcommand (SDSC_DISCON);
      len = receive (buffer, sizeof (buffer), sizeof (buffer), 0);
      return 1;
    }

}

ping ()
{
  byte buf[2048];
  int i;


  sendcommand (SDSC_HELLO);
  buf[0] = 0;
  i = receive (buf, sizeof (buf), 1, 0);
  printf ("%d %x\n", i, buf[0]);

#if 1
    while (1) {

#if 1
  sendcommand (SDSC_GIMME);
  sprintf (buf, "PIC%04d.JPG", 2);

  for (i=0;buf[i];++i ) 
	sendcommand(buf[i]);
 
  sendcommand(0);
#else
      sendcommand (SDSC_START);
#endif

      bzero (buf, sizeof (buf));
      i = receive (buf, sizeof (buf), 133, 0);
      printf ("%d %x %s\n", i, buf[0], buf + 3);

      if (isnullinfo (buf))
        break;

      sendcommand (SDSC_BINARY);
      sendcommand (SDSC_START);
      while (1)
        {
          i = receive (buf, sizeof (buf), 1, 0);
          sendcommand (SDSC_BINARY);

          if (i == 1)
            break;

        }
    }

  //sendcommand (SDSC_BINARY);
#endif

#if 0

  sendcommand (SDSC_GIMME);
  sprintf (buf, "PIC%04d.JPG", 2);

  for (i=0;buf[i];++i ) 
	sendcommand(buf[i]);
 
  sendcommand(0);
  

  bzero (buf, sizeof (buf));
  i = receive (buf, sizeof (buf), 133, 0);
  printf ("%d %x %s\n", i, buf[0], buf + 3);


  sendcommand (SDSC_BINARY);
  sendcommand (SDSC_START);
  while (1)
    {
      i = receive (buf, sizeof (buf), 1, 0);
      sendcommand (SDSC_BINARY);

      if (i == 1)
        break;

    }

  sendcommand (SDSC_BINARY);
#endif

  sendcommand (SDSC_GOODBYE);
  buf[0] = 0;
  i = receive (buf, sizeof (buf), 1, 0);
  printf ("%d %x\n", i, buf[0]);

  sendcommand (SDSC_GOODBYE);
  buf[0] = 0;
  i = receive (buf, sizeof (buf), 1, 0);
  printf ("%d %x\n", i, buf[0]);





}

figure_proto ()
{
  byte buf[2048];
  int i, tries = 3;

  proto = PROTO_PROTO1;
  while (tries--)
    {

//      sendcommand (SDSC_PROTO2);
      buf[0] = 0;
      i = receive (buf, 1, 1, 0);

      if ((i == 1) && (buf[0] == SDSC_PROTO2OK))
        {
          proto = PROTO_PROTO2;
          return;
        }

      if (reset_state ())
        return;
    }

  proto = PROTO_UNKN;

}

/*
   ** Place the camera in remote mode. In this mode, "PC CONNECTED"
   ** will be displayed on the camera's screen.
 */
static int
SDSC_Initialise (void)
{

  fd = open (serial_port, O_RDWR);

  if (fd < 0)
    {
      error_dialog ("Couldn't open camera device");
      return 1;
    }

  if (init_comms (fd))
    {
      error_dialog ("Couldn't init comm port");
      return 1;
    }


/*  figure_proto ();

  switch (proto)
    {
    case PROTO_PROTO1:
      printf ("Using protocol #1\n");
      break;
    case PROTO_PROTO2:
      printf ("Using protocol #2\n");
      break;
    case PROTO_UNKN:
      printf ("Protocol negociation failed\n");
      return 1;
    }



*/
  return 0;

}

/*
   **
 */
static int
SDSC_disconnect ()
{
  unsigned char buffer[2048];
  int len;


#ifdef DEBUG
  fprintf (stderr, "Do disconnect\n");
#endif
#ifdef DEBUG
  fprintf (stderr, "Done disconnect\n");
#endif

  return 0;
}

#ifdef ENABLE_QUERYPICSEX
/*Returns 0 for error or num of pictures in camera */
/*also sets *sizes to sizes in bytes of pictures */

int
SDSC_QueryPicsEx (int *sizes)
{
  unsigned char buffer[2048];
  int len;
  int n;

#ifdef DEBUG
  fprintf (stderr, "Count\n");
#endif

  SDSC_disconnect (fd);

  /*Now at start so count how many... */

  n = 0;

  do
    {
      sendcommand (SDSC_START); /*Read out a header */
      len = receive (buffer, sizeof (buffer), SDSC_INFOSIZE, 1);

      if (len != SDSC_INFOSIZE)
        return 0;

      if (!isnullinfo (buffer))
        {
          int i = 3, s = 0;

          sizes[n] = 0;

          for (; (s < 2) && (i < len); ++i)
            {
              if (!buffer[i])
                s++;
              if (s && isdigit (buffer[i]))
                {
                  sizes[n] *= 10;
                  sizes[n] += buffer[i] - '0';
                }

            }


          n++;
        }

      sendcommand (SDSC_NEXT);

    }
  while (!isnullinfo (buffer));

#ifdef DEBUG
  fprintf (stderr, "Done Count %d\n", n);
#endif

  return n;
}
#endif
/*
   ** returns 0 for error or the number of pictures in the camera
 */
static int
SDSC_QueryPics ()
{
  unsigned char buffer[2048];
  int len;
  int n;

#ifdef DEBUG
  fprintf (stderr, "Count\n");
#endif

  SDSC_disconnect (fd);

  /*Now at start so count how many... */

  n = 0;

  do
    {
      sendcommand (SDSC_START); /*Read out a header */
      len = receive (buffer, sizeof (buffer), SDSC_INFOSIZE, 1);

      if (len != SDSC_INFOSIZE)
        return 0;

      if (!isnullinfo (buffer))
        {
          n++;
        }

      sendcommand (SDSC_NEXT);

    }
  while (!isnullinfo (buffer));

#ifdef DEBUG
  fprintf (stderr, "Done Count %d\n", n);
#endif

  return n;
}

seek_image (int picnum)
{
  unsigned char buffer[2048];
  int i;

  switch (proto)
    {
    case PROTO_PROTO1:
      reset_state ();
      i = picnum;
      while (--i)               /*Seek */
        {
          sendcommand (SDSC_START);
          i = receive (buffer, sizeof (buffer), SDSC_INFOSIZE, 1);
          sendcommand (SDSC_NEXT);
        }
      break;
    case PROTO_PROTO2:
      //  sendcommand (SDSC_DOWNLOAD);
      i = sprintf (buffer, "PIC%04d.JPG", picnum);
      write (fd, buffer, i + 1);
      break;
    }
}


static struct Image *
SDSC_ReadThumb (int picnum)
{
  unsigned char buffer[2048];
  char str[1024];
  int len;


  sendcommand (SDSC_START);
  len = receive (buffer, sizeof (buffer), SDSC_INFOSIZE, 1);
  sendcommand (SDSC_BINARY);
  sendcommand (SDSC_START);

  while (1)
    {
      len = receive (buffer, sizeof (buffer), SDSC_BLOCKSIZE, 1);
    }

}

/*
   ** fill out the rest of the image structure, given "number" already
 */
static struct Image *
SDSC_ReadImage (int picnum, int thumbnail)
{
  unsigned char buffer[2048];
  char str[1024];
  int len;
  int i;
  int blocks;

  struct Image *image = (struct Image *) malloc (sizeof (struct Image));
  byte *ptr;
  int left, size;

#ifdef DEBUG
  fprintf (stderr, "ReadImage \n");
#endif

  if (thumbnail)
    return SDSC_ReadThumb (picnum);

  seek_image (picnum);

  sendcommand (SDSC_START);     /*Read out a header */
  len = receive (buffer, sizeof (buffer), SDSC_INFOSIZE, 1);

  if (len != SDSC_INFOSIZE)
    {
      free (image);
      return 0;
    }

  buffer[SDSC_INFOSIZE] = 0;

  image->image_size = atoi (&buffer[15]);


  sprintf (str, "Length of image %d is %d bytes\n",
           picnum, image->image_size);
  update_status (str);

  blocks = (image->image_size + 0x3ff) / 0x400;
  blocks += 2;

  left = size = 0x400 * blocks;
  i = 0;

  ptr = image->image = (unsigned char *) malloc (size);

  if (!image->image)
    {
      error_dialog ("Failed to malloc image data.");
      free (image);
      return 0;
    }

  update_progress (0.0);

  sendcommand (SDSC_BINARY);    /*Put into image mode */
  sendcommand (SDSC_START);     /*Start reading */

  while (left > 0x400)
    {
      len = receive (buffer, sizeof (buffer), SDSC_BLOCKSIZE, 1);

      if (!len)
        {
          free (image->image);
          free (image);
#ifdef DEBUG
          fprintf (stderr, "ReadImage failed \n");
#endif
          return 0;
        }

      if ((len == 1) && (buffer[1] == 0x4))
        break;

      bcopy (buffer + 3, ptr, 0x400);

      ptr += 0x400;
      left -= 0x400;
      i += 0x400;

      update_progress ((float) i / (float) size);

      sendcommand (SDSC_BINARY); /*Next block */
    }


  if (proto == PROTO_PROTO1)
    reset_state ();

  strcpy (image->image_type, "jpg");
  image->image_info_size = 0;
#ifdef DEBUG
  fprintf (stderr, "Done ReadImage \n");
#endif

  return image;
}


/*
   ** delete an image from the camera
   ** The image number is one based (1 is the first image).
 */
static int
SDSC_DeleteImage (int image_no)
{
  return 0;
}

static char *
SDSC_Summary (void)
{
  return "Not Supported";
}

static char *
SDSC_Description (void)
{
  return "gPhoto Samsung digimax 800k driver\n"
    "(C) James Mckenzie 2000\n" "mailto:james@fishsoup.dhs.org\n";
}

static struct Image *
SDSC_Preview (void)
{
  return 0;
}

static int
SDSC_TakePicture ()
{
  return 0;
}

static int
SDSC_Configure ()
{
  return 0;
}

#if 0
struct _Camera samsung800k = {
  SDSC_Initialise,
  SDSC_ReadImage,
  SDSC_Preview,
  SDSC_DeleteImage,
  SDSC_TakePicture,
  SDSC_QueryPics,
  SDSC_Configure,
  SDSC_Summary,
  SDSC_Description,
};
#endif

main ()
{
  SDSC_Initialise ();

  ping ();

//  SDSC_ReadImage (1, 0);








}
